package pps.core.base.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.io.IoUtil;
import cn.hutool.core.io.NioUtil;
import cn.hutool.core.io.file.FileReader;
import cn.hutool.core.text.CharSequenceUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import org.apache.commons.lang3.StringUtils;
import pps.cloud.base.service.IBaseWeatherCloudService;
import pps.core.base.entity.BasePhotovoltaicPlantView;
import pps.core.base.entity.PlantPredictedPowerDataEnt;
import pps.core.base.enums.WindDirection;
import pps.core.base.mapper.BasePhotovoltaicPlantViewMapper;
import pps.core.base.mapper.PlantPredictedPowerDataMapper;
import pps.core.base.mapper.PlantPredictedPowerDataViewMapper;
import pps.core.base.service.data.DailyData;
import pps.core.base.utils.ReceivedIMAPMail;
import pps.core.common.constant.BusinessConstant;
import pps.core.common.utils.BaseUtils;
import xstartup.annotation.XService;
import xstartup.base.XContext;
import xstartup.base.XStartup;
import xstartup.base.exception.XServiceException;
import xstartup.base.tool.XLoggerTool;
import xstartup.base.util.XCopyUtils;
import xstartup.data.XServiceResult;

import javax.mail.*;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeUtility;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.text.ParseException;
import java.time.LocalTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 天气数据接收Cloud模块
 *
 * @author ZWT
 * @date 2023/11/13 17:54
 */
@XService
public class BaseWeatherCloudServiceImpl implements IBaseWeatherCloudService {

    /**
     * 天气数据接收Cloud模块--天气数据处理
     *
     * @param context 上下文
     * @return {@link XServiceResult}
     */
    @Override
    public XServiceResult weatherDataProcess(XContext context) {
        XLoggerTool logger = context.getLogger();
        String host = XStartup.getCurrent().getProperty("weather.mail.host");
        String username = XStartup.getCurrent().getProperty("weather.mail.username");
        String password = XStartup.getCurrent().getProperty("weather.mail.password");
        String protocol = XStartup.getCurrent().getProperty("weather.mail.protocol");
        //取未读邮件
        Store store = null;
        Folder folder = null;
        List<PlantPredictedPowerDataEnt> deleteList = new ArrayList<>(2048);
        List<PlantPredictedPowerDataEnt> batchList = new ArrayList<>(2048);
        List<String> timeList = Arrays.asList("15:00", "30:00", "45:00");
        DateTime create = DateTime.of(DateUtil.now(), "yyyy-MM-dd HH");
        DateTime end = DateUtil.offsetDay(create, 7);
        try {
            Properties props = new Properties();
            Session session = Session.getDefaultInstance(props, null);
            store = session.getStore(protocol);
            store.connect(host, username, password);
            folder = store.getFolder("INBOX");
            folder.open(Folder.READ_WRITE);
            logger.info("------ weatherDataProcess connect mail ------");
            List<Message> messageList = ReceivedIMAPMail.filterMessage(folder.getMessages(), null, null);
            if (CollUtil.isNotEmpty(messageList)) {
                //取电站
                BasePhotovoltaicPlantViewMapper viewMapper = context.getBean(BasePhotovoltaicPlantViewMapper.class);
                List<BasePhotovoltaicPlantView> plantList = viewMapper.selectPlantList();
                if (CollUtil.isEmpty(plantList)) {
                    logger.info("------ weatherDataProcess plantList is empty ------");
                    return null;
                }
                //按照cityCode分组
                Map<String, List<BasePhotovoltaicPlantView>> plantMap = plantList.stream()
                        .collect(Collectors.groupingBy(BasePhotovoltaicPlantView::getCityCode));
                ReceivedIMAPMail re;
                for (Message message : messageList) {
                    re = new ReceivedIMAPMail((MimeMessage) message);
                    String subject = re.getSubject();
                    if (!CharSequenceUtil.endWith(subject, "天气预测数据")) {
                        continue;
                    }
                    //读取附件
                    re.getMailContent(message);
                    String tempFilePath = this.saveAttachment(message);
                    logger.info("------ weatherDataProcess load file: {}", tempFilePath);
                    FileReader fileReader = new FileReader(tempFilePath);
                    String jsonString = fileReader.readString();
                    JSONObject jsonObject = JSON.parseObject(jsonString);
                    Map<String, List<DailyData>> cityDataMap = new HashMap<>(16);
                    for (Map.Entry<String, Object> entry : jsonObject.entrySet()) {
                        String key = entry.getKey();
                        Object value = entry.getValue();
                        if (value instanceof JSONArray) {
                            cityDataMap.put(key, ((JSONArray) value).toJavaList(DailyData.class));
                        }
                    }
                    for (String cityCode : cityDataMap.keySet()) {
                        if (!plantMap.containsKey(cityCode)) {
                            continue;
                        }
                        List<DailyData> dailyDataList = cityDataMap.get(cityCode);
                        List<BasePhotovoltaicPlantView> plantViewList = plantMap.get(cityCode);
                        //删除数据
                        for (BasePhotovoltaicPlantView plantView : plantViewList) {
                            deleteList.add(PlantPredictedPowerDataEnt.builder()
                                    .plantId(plantView.getId())
                                    .createTime(create)
                                    .endTime(end)
                                    .build());
                        }
                        //遍历预测数据列表
                        for (DailyData dailyData : dailyDataList) {
                            for (BasePhotovoltaicPlantView plant : plantViewList) {
                                List<PlantPredictedPowerDataEnt> allList = new ArrayList<>(128);
                                List<PlantPredictedPowerDataEnt> objList = new ArrayList<>(128);
                                DateTime dateTime = DateTime.of(dailyData.getDateTime());
                                int hours = 0;
                                String plantId = plant.getId();
                                DateTime date;
                                for (int i = 0; i < 8; i++) {
                                    //判断小时数与查到的小时数，如果查到的小时数小，判断为第二天
                                    int hour = LocalTime.parse(dailyData.getHourTimeArray().get(i)).getHour();
                                    if (hours >= hour) {
                                        dateTime = DateUtil.offsetDay(dateTime, 1);
                                    }
                                    hours = hour;
                                    PlantPredictedPowerDataEnt dao = this.getPlantPredictedPowerEnt(plantId,
                                            this.int2Str(hours),
                                            new BigDecimal(CharSequenceUtil.replace(dailyData.getTemperatureArray().get(i), "℃", "")),
                                            new BigDecimal(CharSequenceUtil.replace(dailyData.getWindSpeedArray().get(i), "m/s", "")),
                                            new BigDecimal(Objects.requireNonNull(WindDirection.findByValue(dailyData.getWindDirectionArray().get(i))).getValue()),
                                            new BigDecimal(CharSequenceUtil.replace(dailyData.getPressureArray().get(i), "hPa", "")),
                                            new BigDecimal(CharSequenceUtil.replace(dailyData.getHumidityArray().get(i), "%", "")),
                                            String.valueOf(dateTime.year()),
                                            this.int2Str(dateTime.month() + 1),
                                            this.int2Str(dateTime.dayOfMonth())
                                    );
                                    // 5.执行SQL
                                    allList.add(dao);
                                    PlantPredictedPowerDataEnt isExit = objList.stream().filter(item ->
                                                    CharSequenceUtil.equals(item.getYearTime(), dao.getYearTime()) &&
                                                            CharSequenceUtil.equals(item.getMonthTime(), dao.getMonthTime()) &&
                                                            CharSequenceUtil.equals(item.getDayTime(), dao.getDayTime()) &&
                                                            CharSequenceUtil.equals(item.getHourTime(), dao.getHourTime()))
                                            .findFirst()
                                            .orElse(null);
                                    if (Objects.isNull(isExit)) {
                                        objList.add(dao);
                                    }
                                }
                                // 原站点时间结构为：08:00	11:00	14:00	17:00	20:00	23:00	02:00	05:00
                                // 两条数据中间差了两个小时，使用上下两条数据推算出中间差掉的两个小时
                                for (int i = 0; i < objList.size() - 1; i++) {
                                    for (int y = 1; y < 3; y++) {
                                        PlantPredictedPowerDataEnt dataEnt = objList.get(i);
                                        date = DateTime.of(dataEnt.getDataDate(), "yyyy-MM-dd HH:mm");
                                        String hourString;
                                        Integer hourTime = Integer.valueOf(dataEnt.getHourTime());
                                        if (hourTime.equals(23)) {
                                            date = DateUtil.offsetDay(date, 1);
                                            hourString = "0" + (y - 1);
                                        } else {
                                            if ((hourTime + y) < 10) {
                                                hourString = "0" + (hourTime + y);
                                            } else {
                                                hourString = String.valueOf(hourTime + y);
                                            }
                                        }
                                        PlantPredictedPowerDataEnt nextData = objList.get(i + 1);
                                        BigDecimal valueOf = BigDecimal.valueOf(0.3 * y);
                                        PlantPredictedPowerDataEnt dao = this.getPlantPredictedPowerEnt(plantId,
                                                hourString,
                                                this.compute(dataEnt.getTemperature(), nextData.getTemperature(), valueOf),
                                                this.compute(dataEnt.getWindSpeed(), nextData.getWindSpeed(), valueOf),
                                                dataEnt.getWindDirection(),
                                                this.compute(dataEnt.getPressure(), nextData.getPressure(), valueOf),
                                                this.compute(dataEnt.getHumidity(), nextData.getHumidity(), valueOf),
                                                String.valueOf(date.year()),
                                                this.int2Str(date.month() + 1),
                                                this.int2Str(date.dayOfMonth())
                                        );
                                        // 5.执行SQL
                                        allList.add(dao);
                                    }
                                }
                                //入库
                                List<PlantPredictedPowerDataEnt> sortList = allList.stream()
                                        .sorted(Comparator.comparing(PlantPredictedPowerDataEnt::getDataDate))
                                        .collect(Collectors.toList());
                                for (int i = 0; i < sortList.size(); i++) {
                                    PlantPredictedPowerDataEnt item = sortList.get(i);
                                    batchList.add(item);
                                    if (i < sortList.size() - 1) {
                                        for (int y = 0; y < timeList.size(); y++) {
                                            PlantPredictedPowerDataEnt dao = XCopyUtils.copyNewObject(item);
                                            dao.setMinTime(timeList.get(y));
                                            dao.setDataDate(dao.getYearTime() + '-' + dao.getMonthTime() + '-' + dao.getDayTime() + ' ' + dao.getHourTime() + ':' + dao.getMinTime());
                                            PlantPredictedPowerDataEnt nextData = sortList.get(i + 1);
                                            BigDecimal valueOf = BigDecimal.valueOf(0.3 * (y + 1));
                                            dao.setTemperature(this.compute(Objects.isNull(item.getTemperature()) ? BigDecimal.ZERO : item.getTemperature(),
                                                    Objects.isNull(nextData.getTemperature()) ? BigDecimal.ZERO : nextData.getTemperature(),
                                                    valueOf));
                                            dao.setHumidity(this.compute(Objects.isNull(item.getHumidity()) ? BigDecimal.ZERO : item.getHumidity(),
                                                    Objects.isNull(nextData.getHumidity()) ? BigDecimal.ZERO : nextData.getHumidity(),
                                                    valueOf));
                                            dao.setWindSpeed(this.compute(Objects.isNull(item.getWindSpeed()) ? BigDecimal.ZERO : item.getWindSpeed(),
                                                    Objects.isNull(nextData.getWindSpeed()) ? BigDecimal.ZERO : nextData.getWindSpeed(),
                                                    valueOf));
                                            dao.setPressure(this.compute(Objects.isNull(item.getPressure()) ? BigDecimal.ZERO : item.getPressure(),
                                                    Objects.isNull(nextData.getPressure()) ? BigDecimal.ZERO : nextData.getPressure(),
                                                    valueOf));
                                            dao.setPlaneIrradiance(this.compute(Objects.isNull(item.getPlaneIrradiance()) ? BigDecimal.ZERO : item.getPlaneIrradiance(),
                                                    Objects.isNull(nextData.getPlaneIrradiance()) ? BigDecimal.ZERO : nextData.getPlaneIrradiance(),
                                                    valueOf));
                                            dao.setHorizontalIrradiance(this.compute(Objects.isNull(item.getHorizontalIrradiance()) ? BigDecimal.ZERO : item.getHorizontalIrradiance(),
                                                    Objects.isNull(nextData.getHorizontalIrradiance()) ? BigDecimal.ZERO : nextData.getHorizontalIrradiance(),
                                                    valueOf));
                                            dao.setPower(this.compute(Objects.isNull(item.getPower()) ? BigDecimal.ZERO : item.getPower(),
                                                    Objects.isNull(nextData.getPower()) ? BigDecimal.ZERO : nextData.getPower(),
                                                    valueOf));
                                            batchList.add(dao);
                                        }
                                    }
                                }
                            }
                        }
                    }
                    logger.info("------ weatherDataProcess del temp file path: {}", tempFilePath);
                    FileUtil.del(tempFilePath);
                    message.setFlag(Flags.Flag.DELETED, true);
                }
            }
        } catch (MessagingException e) {
            logger.error("------ weatherDataProcess MessagingException", e);
            throw new XServiceException(e);
        } catch (IOException e) {
            logger.error("------ weatherDataProcess IOException", e);
            throw new XServiceException(e);
        } catch (ParseException e) {
            logger.error("------ weatherDataProcess ParseException", e);
            throw new XServiceException(e);
        } finally {
            logger.info("------ weatherDataProcess finally close");
            NioUtil.close(folder);
            NioUtil.close(store);
        }
        PlantPredictedPowerDataMapper mapper = context.getBean(PlantPredictedPowerDataMapper.class);
        PlantPredictedPowerDataViewMapper dataViewMapper = context.getBean(PlantPredictedPowerDataViewMapper.class);
        int size;
        if (CollUtil.isNotEmpty(deleteList)) {
            size = deleteList.size();
            logger.info("------ weatherDataProcess batchDelete delete plantId: {}", size);
            if (size > BaseUtils.BATCH_SIZE) {
                List<List<PlantPredictedPowerDataEnt>> subList = BaseUtils.getSubList(deleteList);
                for (List<PlantPredictedPowerDataEnt> list : subList) {
                    LambdaQueryWrapper<PlantPredictedPowerDataEnt> wrapper = new LambdaQueryWrapper<>();
                    for (PlantPredictedPowerDataEnt dataEnt : list) {
                        wrapper.or(w -> w.eq(PlantPredictedPowerDataEnt::getPlantId, dataEnt.getPlantId())
                                .between(PlantPredictedPowerDataEnt::getDataDate, dataEnt.getCreateTime(), dataEnt.getEndTime())
                        );
                    }
                    mapper.delete(wrapper);
                }
            } else {
                LambdaQueryWrapper<PlantPredictedPowerDataEnt> wrapper = new LambdaQueryWrapper<>();
                for (PlantPredictedPowerDataEnt dataEnt : deleteList) {
                    wrapper.or(w -> w.eq(PlantPredictedPowerDataEnt::getPlantId, dataEnt.getPlantId())
                            .between(PlantPredictedPowerDataEnt::getDataDate, dataEnt.getCreateTime(), dataEnt.getEndTime())
                    );
                }
                mapper.delete(wrapper);
            }
        }
        if (CollUtil.isNotEmpty(batchList)) {
            size = batchList.size();
            logger.info("------ weatherDataProcess batchList insert size: {}", size);
            if (size > BaseUtils.BATCH_SIZE) {
                List<List<PlantPredictedPowerDataEnt>> subList = BaseUtils.getSubList(batchList);
                subList.forEach(dataViewMapper::insertBatch);
            } else {
                dataViewMapper.insertBatch(batchList);
            }
        }
        return XServiceResult.OK;
    }

    /**
     * 获取电厂预测功率ent
     *
     * @param plantId       电站ID
     * @param hourTime      小时
     * @param temperature   温度
     * @param windSpeed     风速
     * @param windDirection 风向
     * @param pressure      压力
     * @param humidity      湿度
     * @param yearTime      年时间
     * @param monthTime     月份时间
     * @param dayTime       白天
     * @return {@link PlantPredictedPowerDataEnt}
     */
    private PlantPredictedPowerDataEnt getPlantPredictedPowerEnt(String plantId, String hourTime, BigDecimal temperature, BigDecimal windSpeed, BigDecimal windDirection,
                                                                 BigDecimal pressure, BigDecimal humidity, String yearTime, String monthTime, String dayTime) {
        PlantPredictedPowerDataEnt dao = new PlantPredictedPowerDataEnt();
        dao.setCreateTime(DateUtil.date());
        dao.setPlantId(plantId);
        dao.setHourTime(hourTime);
        dao.setTemperature(temperature);
        dao.setWindSpeed(windSpeed);
        dao.setWindDirection(windDirection);
        dao.setPressure(pressure);
        dao.setHumidity(humidity);
        dao.setYearTime(yearTime);
        dao.setMonthTime(monthTime);
        dao.setDayTime(dayTime);
        dao.setMinTime(BusinessConstant.START_OF_DAY_TIME);
        dao.setDataDate(dao.getYearTime() + '-' + dao.getMonthTime() + '-' + dao.getDayTime() + ' ' + dao.getHourTime() + ':' + dao.getMinTime());
        return dao;
    }

    /**
     * 计算
     *
     * @param a 一
     * @param b b
     * @param c c
     * @return {@link BigDecimal}
     */
    private BigDecimal compute(BigDecimal a, BigDecimal b, BigDecimal c) {
        return a.add(b.subtract(a).multiply(c));
    }

    /**
     * int转String
     *
     * @param num 数字
     * @return {@link String}
     */
    private String int2Str(int num) {
        return num < 10 ? "0" + num : String.valueOf(num);
    }

    /**
     * 保存附件
     *
     * @param part 部分
     * @return {@link String}
     * @throws MessagingException 消息传递异常
     * @throws IOException        IOException
     */
    private String saveAttachment(Part part) throws MessagingException, IOException {
        String fileName;
        String filePath = null;
        if (part.isMimeType("multipart/*")) {
            Multipart mp = (Multipart) part.getContent();
            BodyPart mPart;
            for (int i = 0; i < mp.getCount(); i++) {
                mPart = mp.getBodyPart(i);
                String disposition = mPart.getDisposition();
                if (Part.ATTACHMENT.equals(disposition) || Part.INLINE.equals(disposition)) {
                    fileName = mPart.getFileName();
                    if (CharSequenceUtil.isNotEmpty(fileName) && StringUtils.lowerCase(fileName).contains("GB2312")) {
                        fileName = MimeUtility.decodeText(fileName);
                    }
                    fileName = CharSequenceUtil.replace(fileName, "\\?|=", "");
                    filePath = this.saveTempFile(fileName, mPart.getInputStream());
                } else if (mPart.isMimeType("multipart/*")) {
                    this.saveAttachment(mPart);
                } else {
                    fileName = mPart.getFileName();
                    if (CharSequenceUtil.isNotEmpty(fileName) && StringUtils.lowerCase(fileName).contains("GB2312")) {
                        fileName = MimeUtility.decodeText(fileName);
                        filePath = this.saveTempFile(fileName, mPart.getInputStream());
                    }
                }
            }
        } else if (part.isMimeType("message/rfc822")) {
            this.saveAttachment((Part) part.getContent());
        }
        return filePath;
    }

    /**
     * 保存临时文件
     *
     * @param fileName 文件名
     * @param in       在里面
     * @return {@link String}
     */
    private String saveTempFile(String fileName, InputStream in) {
        String storeDir = XStartup.getCurrent().getProperty("weather.file.temp.path");
        StringBuilder stringBuilder = new StringBuilder(storeDir)
                .append(File.separator)
                .append(DateTime.now().toString("yyyyMMdd"));
        String filePath = stringBuilder.append(File.separator).append(fileName).toString();
        FileUtil.touch(filePath);
        FileUtil.writeFromStream(in, filePath);
        IoUtil.close(in);
        return filePath;
    }
}
